-- ============================================================
--
--   The game's "Play" state
--
-- ============================================================


-- ============================================================
-- Define the state and its variables
--   Here will be the actual gameflow
-- ============================================================
--
-- Create the state
local statePlay = clsGame:addState("Play")

-- Current ship with which to play
local objShip = nil

-- List of all in-game asteroids
local tabAsteroids = {}

-- List of all animations currently playing
local tabAnimations = {}
-- print( tabAnimations )
--
-- ============================================================


-- ============================================================
-- Called when entering the state
-- ============================================================
--
function statePlay:enterState()
	-- Generate the ship
	-- print ( "Current ship : " .. tabShipData[currentShip].name )
	local strShipName = tabShipData[currentShip].name
	local strFileName = tabShipData[currentShip].fileName
	local numX, numY  = tabShipData[currentShip].nX, tabShipData[currentShip].nY
	objShip = clsShip:new(strShipName,strFileName, numX, numY)

	-- Generate 10 asteroids
	for i = 1, 10 do
		tabAsteroids[i] = generateAsteroid("asteroid-D00", true)
		tabAsteroids[i].isNew = 0
	end

	-- Reset time for scoring
	elapsedTime = 0
end
--
-- ============================================================


-- ============================================================
-- Called when exiting the state to enter another state
-- ============================================================
--
function statePlay:exitState()
	objShip      = {}
	tabAsteroids = {}
	tabAnimations = {}
	gameScore = math.floor(elapsedTime)
	-- print ( "Elapsed Time : " .. tostring(elapsedTime) )
	-- print ( "Score        : " .. tostring(gameScore  ) )
end
--
-- ============================================================


-- ============================================================
-- update() callback
--   called instead of the regular clsGame:update()
--   when the game is in the Menu state
-- ============================================================
--
function statePlay:update(dt)
	-- DEBUG
	-- print( string.format( "FPS = %6.3f ; %i Asteroids ; %i Animations", 1/dt, #tabAsteroids, #tabAnimations ) )

	-- Get # of asteroids
	local numNumberOfAsteroids = #tabAsteroids

	-- Interprets commands
	local cmdRotation, cmdAcceleration = interpretCommands()

	-- Apply commands on ship
	objShip:update(dt, cmdRotation, cmdAcceleration)

	-- Move asteroids
	for i=1,#tabAsteroids do
		tabAsteroids[i]:update(dt)
	end

	-- Check asteroids collision to ship : end game
	for i=1,#tabAsteroids do
		-- Check if asteroid collided or is about to collide with the player's ship
		tabAsteroids[i].isCollision, tabAsteroids[i].isDanger = tabAsteroids[i]:hasCollidedWith(objShip)

		-- If collision, then exit to GameOver state
		if tabAsteroids[i].isCollision then
			objGame:gotoState("GameOver")
			break
		end
	end

	-- Check asteroids collision between themselves : destroy both asteroids involved
	-- Only mark asteroids to be destroyed, as destroying them in the middle of the loop
	-- would wreck havoc in the loop...
	for i=1,#tabAsteroids do
		for j=i+1,#tabAsteroids do
			-- Check if asteroid collided or is about to collide with the player's ship
			if tabAsteroids[i]:hasCollidedWith(tabAsteroids[j]) then
				-- Mark both asteroids as to be destroyed
				tabAsteroids[i].isAlive = false
				tabAsteroids[j].isAlive = false
				-- Create new asteroids from the collision
				createAsteroidsFromCollision(tabAsteroids, tabAnimations, i, j, "asteroid-D00")
			end
		end
	end

	-- Mark for deletion asteroids out of the screen and whose speed make them move away from the screen
	for i=1,#tabAsteroids do
		local objAsteroid = tabAsteroids[i]
		if (
			   (objAsteroid.position.x<-100                         and objAsteroid.speed.x<0) or (objAsteroid.position.y<-100                          and objAsteroid.speed.y<0)
			or (objAsteroid.position.x>love.graphics.getWidth()+100 and objAsteroid.speed.x>0) or (objAsteroid.position.x>love.graphics.getHeight()+100 and objAsteroid.speed.y>0)
		) then
			tabAsteroids[i].isAlive = false
		end
	end

	-- Actually destroy marked asteroids
	for k=#tabAsteroids,1,-1 do
		if not tabAsteroids[k].isAlive then
			destroyAsteroid(tabAsteroids,k)
		end
	end

	-- Actually destroy finished animations
	-- print( tabAnimations )
	for k=#tabAnimations,1,-1 do
		if tabAnimations[k].currentFrame==-1 then
			destroyAnimation(tabAnimations,k)
		end
	end

	-- Generate an extra asteroid outside the screen if timer expires
	local spawnTimerInterval = 1
	if spawnTimer > spawnTimerInterval then
		tabAsteroids[#tabAsteroids+1] = generateAsteroid("asteroid-D00", false)
		spawnTimer = 0
	end
end
--
-- ============================================================


-- ============================================================
-- draw() callback
--   called instead of the regular clsGame:draw()
--   when the game is in the Menu state
-- ============================================================
--
function statePlay:draw()
	-- Draw the asteroids below ...
	for i=1,#tabAsteroids do
		tabAsteroids[i]:draw()
	end

	-- ... then the animations ...
	for i=1,#tabAnimations do
		tabAnimations[i]:draw()
	end

	-- ... then the ship above
	objShip:draw()
	objShip:drawCollisionCircle()
end
--
-- ============================================================


-- ============================================================
-- keypressed and keyreleased callback
--   called instead of the regular clsGame:keyreleased()
--   when the game is in the Menu state
-- ============================================================
--
function statePlay:keypressed(key)   -- we do not need the unicode, so we can leave it out
	-- DEBUG
	-- print ( "State : Play" )
	-- print("Key : \"" .. tostring(key) .. "\" has been pressed")
	if key=="escape" then
		objGame:gotoState("Menu")
	end
end
function statePlay:keyreleased(key)
	-- DEBUG
	-- print("Key : \"" .. tostring(key) .. "\" has been released")
end
--
-- ============================================================

