-- ============================================================
--
--   All Space Object Class
--
-- ============================================================


-- ============================================================
--   Class definition
-- ============================================================
--
clsSpaceObject = class("clsSpaceObject")
--
-- ============================================================


-- ============================================================
--   initialize : called at new()
-- ============================================================
--
function clsSpaceObject:initialize(
	strFileName, numNX, numNY, numScale,
	numX, numY, numHeading,
	numSpeedX, numSpeedY, numRotationSpeed
)
	-- DEBUG
	-- print ( strFileName, numNX, numNY, numScale, numX, numY, numHeading, numSpeedX, numSpeedY, numRotationSpeed )

	-- Check parameters and enforce default values
	if strFileName == "" or strFileName == nil then
		strFileName = "Alaith"
	end
	if numNX == nil then
		numNX = 6
	end
	if numNY == nil then
		numNY = 6
	end
	if numScale == nil then
		numScale = 1
	end
	if numX == nil then
		numX = love.graphics.getWidth()/2
	end
	if numY == nil then
		numY = love.graphics.getHeight()/2
	end
	if numHeading == nil then
		numHeading = 0
	end
	if numSpeedX == nil then
		numSpeedX = 0
	end
	if numSpeedY == nil then
		numSpeedY = 0
	end
	if numRotationSpeed == nil then
		numRotationSpeed = 0
	end

	-- DEBUG
	-- print ( strFileName, numNX, numNY, numX, numY, numHeading, numSpeedX, numSpeedY, numRotationSpeed )

	-- State properties
	self.position       = {x = numX     , y = numY     , heading  = numHeading      }
	self.speed          = {x = numSpeedX, y = numSpeedY, rotation = numRotationSpeed}
	self.acceleration   = {x =         0, y =         0}

	-- DEBUG
	-- print ( self.speed.x, self.speed.y, self.speed.rotation )

	-- Sprite property
	self.sprite = {
		image  = love.graphics.newImage(strFileName),
		number = {x = numNX, y = numNY},
		scale = numScale,
		frame  = {}
	}
	numImageWidth, numImageHeight = self.sprite.image:getDimensions()
	self.sprite.width  = numImageWidth/self.sprite.number.x
	self.sprite.height = numImageHeight/self.sprite.number.y
	self.sprite.total  = self.sprite.number.x * self.sprite.number.y

	-- Generate the quads to display the images in the sprite
	-- frame[i] corresponds to [i, i+10[
	for y=0, self.sprite.number.y-1, 1 do
		for x=0, self.sprite.number.x-1, 1 do
			i = x+y*self.sprite.number.x
			self.sprite.frame[i] = love.graphics.newQuad(
				x*self.sprite.width,
				y*self.sprite.height,
				self.sprite.width,
				self.sprite.height,
				self.sprite.image:getDimensions())
		end
	end

	-- Flags
	self.isNew       = 50    -- # of frames to display a new object
	self.isAlive     = true  -- Object is alive, if false, must be destroyed
	self.isDanger    = false -- Object is not yet dangerous for the player
	self.isCollision = false -- Object has not yet collided with player
end
--
-- ============================================================


-- ============================================================
--   update : called by the game's global love.update()
-- ============================================================
--
function clsSpaceObject:update(dt)
	-- Update speed in pixel/sec from acceleration
	self.speed.x = self.speed.x + dt*self.acceleration.x
	self.speed.y = self.speed.y + dt*self.acceleration.y

	-- Update position in pixel from speed
	self.position.x = self.position.x + dt*self.speed.x
	self.position.y = self.position.y + dt*self.speed.y

	-- Update heading in degrees from rotation speed
	self.position.heading = (self.position.heading + dt*self.speed.rotation) % math.rad(360)

	-- DEBUG
	-- print ( string.format ( "Position  : (pos.x : %f ; pos.y : %f ; heading : %f", self.position.x, self.position.y, math.deg(self.position.heading) ) )
end
--
-- ============================================================


-- ============================================================
--   draw : called by the game's global love.draw()
-- ============================================================
--
function clsSpaceObject:draw()
	if self.isNew>0 then
		-- Draw new objects in green
		-- love.graphics.setColor(0,255,0)
		self.isNew = self.isNew-1
	end

	if self.sprite.total == 1 then
		self.drawSpriteMonoframe(self)
	else
		self.drawSpriteMultiframe(self)
	end

	-- Return to default drawing color
	love.graphics.setColor(255,255,255)

	-- If the object is a danger, draw speed and circle in orange
	if self.isDanger then
		love.graphics.setColor(255,127,0)
	end
	-- If the object is colliding, draw in red
	if self.isCollision then
		love.graphics.setColor(255,0,0)
	end

	-- Always draw the speed vector
	self.drawSpeedVector(self)

	-- Draw collision circle only if danger and/or colliding
	if self.isDanger or self.isCollision then
		self:drawCollisionCircle()
	end

	-- Return to default drawing color
	love.graphics.setColor(255,255,255)
end
--
-- ============================================================


-- ============================================================
--   Draw sprite only for multiframe objects : select frame
-- ============================================================
--
function clsSpaceObject:drawSpriteMultiframe()
	-- Get current frame
	currentFrame = ( self.sprite.total - math.floor( math.deg(self.position.heading) / (360/self.sprite.total) ) ) % self.sprite.total

	-- Get corresponding quad for selecting sprite
	activeFrame = self.sprite.frame[currentFrame]

	-- Draw the sprite
	love.graphics.draw(
		self.sprite.image,
		activeFrame,
		self.position.x,
		self.position.y,
		0,
		self.sprite.scale,
		self.sprite.scale,
		self.sprite.width/2,
		self.sprite.height/2
	)
end
--
-- ============================================================


-- ============================================================
--   Draw sprite only for one-frame objects : rotate image
-- ============================================================
--
function clsSpaceObject:drawSpriteMonoframe()
	--DEBUG
	-- print ( "clsSpaceObject:drawSpriteMonoframe : ", math.deg(self.position.heading) )

	-- Draw the sprite
	love.graphics.draw(
		self.sprite.image,
		self.sprite.frame[0],
		self.position.x,
		self.position.y,
		self.position.heading,
		self.sprite.scale,
		self.sprite.scale,
		self.sprite.width/2,
		self.sprite.height/2
	)
end
--
-- ============================================================


-- ============================================================
--   Draw Speed vector on object
-- ============================================================
--
function clsSpaceObject:drawSpeedVector()
	-- Draw the speed vector
	love.graphics.line(
		self.position.x,
		self.position.y,
		self.position.x + self.speed.x,
		self.position.y + self.speed.y
	)
end
--
-- ============================================================


-- ============================================================
--   Draw Collision Circle on object
-- ============================================================
--
function clsSpaceObject:drawCollisionCircle()
	-- Compute collision circle
	local numDiameter = self.sprite.width
	if self.sprite.width < self.sprite.height then
		numDiameter = self.sprite.height
	end
	numDiameter = numDiameter*self.sprite.scale

	-- Then draw it
	love.graphics.circle(
		"line",
		self.position.x,
		self.position.y,
		numDiameter/2
	)
end
--
-- ============================================================


-- ============================================================
-- function to compute collision between two objects
-- ============================================================
--
function clsSpaceObject:hasCollidedWith(objOtherSpaceObject)
	-- Declare local variable
	local numDistance     = 0
	local boolIsDanger    = false
	local boolIsCollision = false

	-- If asteroid has just been created,no collision
	if self.isNew>0 or objOtherSpaceObject.isNew>0 then
		return false, false
	end

	-- Compute collision diameters
	local numSelfDiameter = self.sprite.width
	if self.sprite.width < self.sprite.height then
		numSelfDiameter = self.sprite.height
	end
	numSelfDiameter = numSelfDiameter*self.sprite.scale
	local numOtherDiameter = objOtherSpaceObject.sprite.width
	if objOtherSpaceObject.sprite.width < objOtherSpaceObject.sprite.height then
		numOtherDiameter = objOtherSpaceObject.sprite.height
	end
	numOtherDiameter = numOtherDiameter*objOtherSpaceObject.sprite.scale

	-- For danger flag, use distance in 1 sec, taking both speeds into account
	numDistance = math.sqrt(
		  (objOtherSpaceObject.position.x+objOtherSpaceObject.speed.x-self.position.x-self.speed.x)^2
		+ (objOtherSpaceObject.position.y+objOtherSpaceObject.speed.y-self.position.y-self.speed.y)^2
	)
	local numNormalizedDistance = numDistance / ((numSelfDiameter + numOtherDiameter)/2)
	boolIsDanger = (numNormalizedDistance<1.2)

	-- For collision flag, use instant positions
	numDistance = math.sqrt(
		  (objOtherSpaceObject.position.x-self.position.x)^2
		+ (objOtherSpaceObject.position.y-self.position.y)^2
	)
	local numNormalizedDistance = numDistance / ((numSelfDiameter + numOtherDiameter)/2)
	boolIsCollision = (numNormalizedDistance<1)

	return boolIsCollision, boolIsDanger
end
--
-- ============================================================
